@@ -49,23 +49,25 @@ or by creating a row in DynamoDB to lock the file. The table in DynamoDB is "new
 
 """
 
-import time
-import json
 import datetime
+import hashlib
+import json
 import logging
 import os
-import gzip as gz
 import re
 import shutil
-import hashlib
+import time
+from typing import Optional, List, Tuple, Iterator, Union, Any, Type
 
 from newtools.aws import S3Location
-from .sql_client import SqlClient
 from .athena import AthenaClient
+from .sql_client import SqlClient
 from newtools.doggo import DoggoLock, DoggoFileSystem, DynamoDogLock, FileDoggo, PandasDoggo
 
 # Set S3fsMixin for backwards compatibility
 S3fsMixin = DoggoFileSystem
+Path = Union[str, S3Location]
+DataFrame = Type['pd.DataFrame']
 
 
 class BaseCachedClass:
@@ -126,6 +128,7 @@ class BaseCachedClass:
         :param files_to_hash: the name of the file to hash in the output name
         :param data_hash: an additional hash to add to the string
         :param suffix: a suffix to add to the end
+
         :return: the full path to cache to
         """
 
@@ -143,14 +146,22 @@ class BaseCachedClass:
             epoch = int(int(time.time()))
             hash_list.append(str(epoch - epoch % self._expiry_seconds))
 
-        file = "_".join(hash_list) + ".csv" + suffix
+        file_format = self.get_cache_file_format(suffix)
+        file_loc = "_".join(hash_list) + file_format
 
         if self.cache_path.startswith("s3://"):
-
-            return S3Location(self.cache_path).join(file)
-
+            return S3Location(self.cache_path).join(file_loc)
         else:
-            return os.path.join(self.cache_path, file)
+            return os.path.join(self.cache_path, file_loc)
+
+    @staticmethod
+    def get_cache_file_format(
+        suffix: str
+    ) -> str:
+        """
+        Return the file format for the cache path.
+        """
+        return ".csv" + suffix
 
 
 class BaseCachedQuery(BaseCachedClass):
@@ -330,12 +341,84 @@ class BaseCachedQuery(BaseCachedClass):
         """
         pass
 
-    def get_results(self,
-                    sql_file,
-                    output_prefix,
-                    params=None,
-                    replacement_dict=None,
-                    refresh_cache=False):
+    def get_query_params(
+        self,
+        params: Optional[dict],
+        ignore_list: List[str]
+    ) -> dict:
+        """
+        Get a set of query parameters, with passed arguments taking precedence.
+        Also formats any list parameters as per __format_list().
+
+        :param ignore_list: any parameters to ignore from list formatting
+        """
+        query_parameters = dict(self._args)
+        if params is not None:
+            query_parameters.update(params)
+
+        for arg in query_parameters:
+            if type(query_parameters[arg]) == list and arg not in ignore_list:
+                query_parameters[arg] = self.__format_list(query_parameters[arg])
+
+        return query_parameters
+
+    def check_exists(
+        self,
+        output_file: str,
+        refresh_cache: bool
+    ) -> Optional[str]:
+        """
+        Function to check if the output file already exists in S3.
+        If so, return the file location or remove the file if opting to refresh the cache.
+        """
+        if self._exists(output_file):
+            if refresh_cache:
+                self._logger.info("Refreshing cache at {0}".format(output_file))
+                with self._get_lock(output_file):
+                    self.dfs.rm(output_file, recursive=True)
+            else:
+                self._logger.info("Loading query from {0}".format(output_file))
+                return output_file
+
+    def lock_and_execute(
+        self,
+        output_file: str,
+        file_path: str,
+        query_parameters: dict,
+        replacement_dict: dict
+    ) -> None:
+        """
+        Function to obtain the lock on the output file path and execute the query.
+        """
+        with self._get_lock(output_file):
+            if not self._exists(output_file):
+                # clean up any S3 files with the same prefix
+                for file in self.dfs.glob(output_file + "**"):  # pragma: nocover
+                    self.dfs.rm(file)
+
+                self._logger.info("Executing query to {0}".format(output_file))
+                self._execute_query(
+                    query_file=file_path,
+                    output_file=output_file,
+                    query_parameters=query_parameters,
+                    replacement_dict={} if replacement_dict is None else replacement_dict
+                )
+
+                # In the S3 case, redshift files are saved with a prefix and need to be renamed
+                if self._redshift_checks and output_file.startswith("s3://"):
+                    if not self.dfs.exists(output_file):
+                        files = self.dfs.glob(output_file + "**")
+                        self._concatenate_files(output_file=output_file, files=files)
+
+    def get_results(
+        self,
+        sql_file: str,
+        output_prefix: str,
+        params: Optional[dict] = None,
+        replacement_dict: Optional[dict] = None,
+        refresh_cache: Optional[bool] = False,
+        ignore_list: Optional[List[str]] = None
+    ) -> str:
         """
         Runs the specified SQL file
 
@@ -344,20 +427,11 @@ class BaseCachedQuery(BaseCachedClass):
         :param params: the parameters to use in the query
         :param replacement_dict: and replacement to be made in the query's text
         :param refresh_cache: if set, then the cache for this query is refreshed when it is run
-        :return:
+        :param ignore_list: any parameters we do not want to __format_list on
         """
+        ignore_list = ignore_list if ignore_list else []
+        query_parameters = self.get_query_params(params, ignore_list)
 
-        # get a set of query parameters, with passed arguments taking precedence
-        query_parameters = dict(self._args)
-        if params is not None:
-            query_parameters.update(params)
-
-        # Format lists
-        for arg in query_parameters:
-            if type(query_parameters[arg]) == list:
-                query_parameters[arg] = self.__format_list(query_parameters[arg])
-
-        # set the path
         file_path = self.get_sql_file(sql_file)
         self._validate_sql(file_path)
 
@@ -365,38 +439,21 @@ class BaseCachedQuery(BaseCachedClass):
             prefix=output_prefix,
             files_to_hash=[file_path],
             data_hash=self._get_dict_hash(params, replacement_dict),
-            suffix=".gz" if self._gzip else "")
+            suffix=".gz" if self._gzip else ""
+        )
 
         self._log_parameters(params)
 
-        if self._exists(output_file):
-            if refresh_cache:
-                self._logger.info("Refreshing cache at {0}".format(output_file))
-                with self._get_lock(output_file):
-                    self.dfs.rm(output_file)
-            else:
-                self._logger.info("Loading query from {0}".format(output_file))
-                return output_file
+        cached_loc = self.check_exists(output_file, refresh_cache)
+        output_file = cached_loc if cached_loc else output_file
 
-        # Get the lock and create the output file
-        with self._get_lock(output_file):
-            if not self._exists(output_file):
-                # clean up any S3 files with the same prefix
-                for file in self.dfs.glob(output_file + "**"):  # pragma: nocover
-                    self.dfs.rm(file)
+        self.lock_and_execute(
+            output_file=output_file,
+            file_path=file_path,
+            query_parameters=query_parameters,
+            replacement_dict=replacement_dict
+        )
 
-                self._logger.info("Executing query to {0}".format(output_file))
-                self._execute_query(query_file=file_path,
-                                    output_file=output_file,
-                                    query_parameters=query_parameters,
-                                    replacement_dict={} if replacement_dict is None else replacement_dict)
-
-                # In the S3 case, redshift files are saved with a prefix and need to be renamed
-                if self._redshift_checks and output_file.startswith("s3://"):
-                    if not self.dfs.exists(output_file):
-                        files = self.dfs.glob(output_file + "**")
-                        self._concatenate_files(output_file=output_file,
-                                                files=files)
         return output_file
 
     def _concatenate_files(self, output_file, files, deleting_files=True):
@@ -516,7 +573,7 @@ class CachedAthenaQuery(BaseCachedQuery):
     """
     The cached AthenaQuery will execute queries against AWS Athena and cache results in csv.gz format.
 
-    It takes single additiona arguments, queue_queries, which defaults to False but will queue the queries
+    It takes single additional arguments, queue_queries, which defaults to False but will queue the queries
     before executing them if set to true.
 
     Two special parameters must be passed to connect to Athena:
@@ -584,11 +641,30 @@ class CachedAthenaQuery(BaseCachedQuery):
             f.write('-- Parameters: {0}\n'.format(self._clean_dict(parameters)).encode('UTF-8'))
             f.write((logged_query + ';\n').encode('UTF-8'))
 
-    def _execute_query(self,
-                       query_file,
-                       output_file,
-                       query_parameters,
-                       replacement_dict):
+    @staticmethod
+    def get_query_body(
+        query_file: str,
+        query_parameters: dict,
+        replacement_dict: dict
+    ) -> str:
+        """
+        Function to open the query and format with the specified parameters.
+        """
+        with open(query_file) as f:
+            actual_query = f.read()
+
+        for key in replacement_dict:
+            actual_query = actual_query.replace(key, replacement_dict[key])
+
+        return actual_query.format(**query_parameters)
+
+    def _execute_query(
+        self,
+        query_file: str,
+        output_file: str,
+        query_parameters: dict,
+        replacement_dict: dict
+    ) -> None:
         """
         Executes the query
 
@@ -596,27 +672,15 @@ class CachedAthenaQuery(BaseCachedQuery):
         :param output_file: the location to store the output
         :param query_parameters: the full set of query parameters to use
         :param replacement_dict: any items to replace directly in the SQL code
-        :return:
         """
-
-        # run the query to unload them
-        with open(query_file) as f:
-            actual_query = f.read()
-
-            # Apply any replacements
-            for key in replacement_dict:
-                actual_query = actual_query.replace(key, replacement_dict[key])
-
-            # Apply the parameters
-            actual_query = actual_query.format(
-                **query_parameters)
+        actual_query = self.get_query_body(query_file, query_parameters, replacement_dict)
 
         # Save an archive copy of the query
         archive_query_path = self.archive_path(query_file)
         if archive_query_path is not None:
             self._archive_query(actual_query, query_parameters, archive_query_path)
 
-        # Now now the query on Athena
+        # Now run the query on Athena
         self._results[output_file] = self._ac.add_query(
             sql=actual_query,
             name="unload {}".format(output_file),
@@ -642,3 +706,227 @@ class CachedAthenaQuery(BaseCachedQuery):
             # delete the metadata file
             self.dfs.rm(actual_file)
             self.dfs.rm(actual_file + '.metadata')
+
+
+class CachedCTASQuery(CachedAthenaQuery):
+    """
+    Class allowing functionality to run Athena queries using the CTAS operation.
+    """
+    def __init__(
+        self,
+        params: Optional[dict] = None,
+        cache_path: Optional[Path] = '',
+        sql_archive_path: Optional[str] = None,
+        sql_paths: Optional[Union[str, List[str]]] = None,
+        dynamodb_lock: Optional[bool] = True,
+        logger: Optional[logging.Logger] = logging.getLogger(__name__),
+        expiry_seconds: Optional[int] = 0
+    ) -> None:
+        if cache_path and not cache_path.startswith('s3://'):
+            raise AssertionError(f"CTAS queries can only output to S3; current cache path: {cache_path}")
+
+        super().__init__(
+            params=params,
+            cache_path=cache_path,
+            sql_archive_path=sql_archive_path,
+            sql_paths=sql_paths,
+            dynamodb_lock=dynamodb_lock,
+            logger=logger,
+            expiry_seconds=expiry_seconds
+        )
+        self.doggo = PandasDoggo()
+
+    @classmethod
+    def get_query_header(
+        cls,
+        database: str,
+        table_name: str,
+        external_location: Path,
+        output_format: Optional[str] = None,
+        output_compression: Optional[str] = None,
+        partition_columns: Optional[Union[str, List[str]]] = None,
+        bucket_info: Optional[Tuple[str, int]] = None
+    ) -> str:
+        """
+        Function to calculate the CTAS header for the query to run.
+
+        :param database: the database to CTAS from
+        :param table_name: the name of the CTAS temporary table
+        :param external_location: where to save the CTAS contents to in S3
+        :param output_format: the format of the output data, which is PARQUET by default
+        :param output_compression: the compression of the output data
+            https://docs.aws.amazon.com/athena/latest/ug/compression-formats.html
+        :param partition_columns: the columns to partition by
+            must ensure the partition columns are the last columns of the SELECT statement in the query
+            moreover, if multiple partition columns are specified, their ordering must be as in the query
+        :param bucket_info: a tuple containing the column to bucket on and the bucket count
+        """
+        base_template = "CREATE TABLE {db}.{table} WITH({query_prop}) AS"
+
+        properties = {
+            "format": f"'{output_format if output_format else 'PARQUET'}'",
+            "write_compression": f"'{output_compression if output_compression else 'SNAPPY'}'",
+            "external_location": f"'{external_location}'"
+        }
+
+        if partition_columns:
+            if isinstance(partition_columns, str):
+                partition_columns = [partition_columns]
+            properties["partitioned_by"] = "ARRAY{}".format(partition_columns)
+        if bucket_info:
+            column, count = bucket_info
+            properties["bucketed_by"] = "ARRAY{}".format([column])
+            properties["bucket_count"] = int(count)
+
+        str_properties = ", ".join([f"{key} = {properties[key]}" for key in properties])
+
+        return base_template.format(
+            db=database,
+            table=table_name,
+            query_prop=str_properties
+        )
+
+    def get_results(
+        self,
+        sql_file: str,
+        output_prefix: str,
+        params: Optional[dict] = None,
+        replacement_dict: Optional[dict] = None,
+        refresh_cache: Optional[bool] = False,
+        **kwargs
+    ) -> str:
+        """
+        Function to run the specified query and return the output path(/directory) to the data.
+        """
+        return super().get_results(
+            sql_file=sql_file,
+            output_prefix=output_prefix,
+            params=params,
+            replacement_dict=replacement_dict,
+            refresh_cache=refresh_cache,
+            ignore_list=['partition_columns']
+        )
+
+    @staticmethod
+    def get_cache_file_format(*args, **kwargs) -> str:
+        """
+        Return the file format for the cache path for CTAS.
+        """
+        return "/"
+
+    def _execute_query(
+        self,
+        query_file: str,
+        output_file: str,
+        query_parameters: dict,
+        replacement_dict: dict
+    ) -> None:
+        """
+        Function to execute the query under CTAS.
+        """
+        self.validate_ctas_params(query_parameters)
+        ctas_params = self.extract_ctas_parameters(query_parameters, output_file)
+        ctas_header = self.get_query_header(**ctas_params)
+
+        query_body = self.get_query_body(query_file, query_parameters, replacement_dict)
+
+        actual_query = f"{ctas_header}\n{query_body}"
+
+        # Save an archive copy of the query
+        archive_query_path = self.archive_path(query_file)
+        if archive_query_path is not None:
+            self._archive_query(actual_query, query_parameters, archive_query_path)
+
+        # Run the query on Athena
+        self._ac.add_query(
+            sql=actual_query,
+            name="CTAS {}".format(output_file),
+            output_location=self.cache_path
+        )
+        self._ac.wait_for_completion()
+
+        # Drop the CTAS table
+        self._ac.add_query(
+            sql=f"DROP TABLE IF EXISTS {ctas_params['database']}.{ctas_params['table_name']}",
+            name="DROP CTAS {}".format(output_file),
+            output_location=self.cache_path
+        )
+        self._ac.wait_for_completion()
+
+    @staticmethod
+    def validate_ctas_params(
+        parameters: dict
+    ) -> None:
+        """
+        Function to check the required parameters for CTAS are specified in the parameters.
+        """
+        required_params = ["ctas_athena_db"]
+
+        for param in required_params:
+            assert param in parameters, f"Required CTAS parameter '{param}' has not been specified"
+
+    @staticmethod
+    def extract_ctas_parameters(
+        params: dict,
+        output_file: str
+    ) -> dict:
+        """
+        Function to extract the CTAS parameters from the query parameters.
+
+        :param params: the query parameters and generic SQL parameters
+        :param output_file: the external location for CTAS, of the form {cache_path}/{sql_hash}/
+        """
+        return {
+            "database": params.get("ctas_athena_db"),
+            "table_name": output_file.split('/')[-2],
+            "external_location": output_file,
+            "output_format": params.get("output_format"),
+            "output_compression": params.get("output_compression"),
+            "partition_columns": params.get("partition_columns"),
+            "bucket_info": params.get("bucket_info")
+        }
+
+    def load_dir_iter(
+        self,
+        load_path: Path,
+        file_format: Optional[str] = None
+    ) -> Iterator[DataFrame]:
+        """
+        Function to return a generator of the output CTAS data.
+
+        :param load_path: the directory containing the CTAS data in S3
+        :param file_format: the format of the save files, e.g. PARQUET or CSV
+        """
+        yield from self.doggo.load_dir_iter(load_path, file_format)
+
+    def load_dir(
+        self,
+        load_path: Path,
+        file_format: Optional[str] = None
+    ) -> DataFrame:
+        """
+        Function to load in the CTAS output and return one unified dataframe.
+
+        :param load_path: the directory containing the CTAS data in S3
+        :param file_format: the format of the save files, e.g. PARQUET or CSV
+        """
+        return self.doggo.load_dir(load_path, file_format)
+
+    def get_partitions(
+        self,
+        load_path: Path
+    ) -> List[Tuple[Any]]:
+        """
+        Function to load all partitions under the load_path into a list of tuples of the form:
+            ('partition_col1=partition_val1', 'partition_col2=partition_val2', ..., '{path_to_data}')
+
+        :param load_path: the directory containing the CTAS data in S3
+        """
+        output: List[tuple] = []
+
+        for file_name in self.dfs.ls(location=load_path, recursive=True):
+            partitions = tuple(ele for ele in file_name.split('/') if "=" in ele)
+            external_location = re.search('(.*=.*/).*', file_name).group(1)
+            output.append(partitions + (external_location, ))
+
+        return list(set(output))
